// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#include "veins/modules/application/lanradio/SimpleApp.h"
#include <chrono>

Define_Module(SimpleApp);

const simsignalwrap_t SimpleApp::mobilityStateChangedSignal = simsignalwrap_t(MIXIM_SIGNAL_MOBILITY_CHANGE_NAME);
const simsignalwrap_t SimpleApp::parkingStateChangedSignal = simsignalwrap_t(TRACI_SIGNAL_PARKING_CHANGE_NAME);

void SimpleApp::initialize(int stage) {
    BaseApplLayer::initialize(stage);

    if (stage==0) {

        //initialize pointers to other modules
        if (FindModule<TraCIMobility*>::findSubModule(getParentModule())) {
            mobility = TraCIMobilityAccess().get(getParentModule());
            traci = mobility->getCommandInterface();
            traciVehicle = mobility->getVehicleCommandInterface();
        }
        else {
            traci = NULL;
            mobility = NULL;
            traciVehicle = NULL;
        }

        annotations = AnnotationManagerAccess().getIfExists();
        ASSERT(annotations);

        mac = FindModule<WaveAppToMac1609_4Interface*>::findSubModule(
                getParentModule());
        assert(mac);

        myId = getParentModule()->getId();
        sent_cams_max_size = 4;
        sent_spats_max_size = 4;

        //read parameters
        headerLength = par("headerLength").longValue();
        beaconLengthBits = par("beaconLengthBits").longValue();
        beaconPriority = par("beaconPriority").longValue();

        communicateWhileParked = par("communicateWhileParked").boolValue();
        app_debug = par("app_debug").boolValue();
        sendBeacons = par("sendBeacons").boolValue();

        isParked = false;


        findHost()->subscribe(mobilityStateChangedSignal, this);
        findHost()->subscribe(parkingStateChangedSignal, this);

        first_CAM_generated = false;
        first_SPAT_generated = false;


        // MAC Address
        int a = 0xF2; //intuniform(0,0xFF) | 0x2;
        int b = intuniform(0,0xFF);
        int c = intuniform(0,0xFF);
        int d = intuniform(0,0xFF);
        int e = intuniform(0,0xFF);
        int f = intuniform(0,0xFF);

        //std::stringstream srcmac;
        char srcMac[18];
        sprintf(srcMac, "%x:%x:%x:%x:%x:%x", a, b, c, d, e, f);
        myMacAddress = srcMac;

        //std::cerr << "myMacAddress: " << myMacAddress << std::endl;

    }
    else if (stage == 1) {

    }
}

void SimpleApp::populateWSM(WaveShortMessage* wsm, int rcvId, int serial) {

    wsm->setWsmVersion(1);
    wsm->setTimestamp(simTime());
    wsm->setSenderAddress(myId);
    wsm->setRecipientAddress(rcvId);
    wsm->setSerial(serial);
    wsm->setBitLength(headerLength);


    if (BasicSafetyMessage* bsm = dynamic_cast<BasicSafetyMessage*>(wsm) ) {
        bsm->setSenderPos(curPosition);
        bsm->setSenderPos(curPosition);
        bsm->setSenderSpeed(curSpeed);
        bsm->setPsid(-1);
        bsm->setChannelNumber(Channels::CCH);
        bsm->addBitLength(beaconLengthBits);
        wsm->setUserPriority(beaconPriority);
    }

    std::stringstream ss;
    ss << "this is a message from node: " << getParentModule()->getFullPath();
    wsm->setWsmData(ss.str().c_str());
}

void SimpleApp::receiveSignal(cComponent* source, simsignal_t signalID, cObject* obj, cObject* details) {
    Enter_Method_Silent();
    if (signalID == mobilityStateChangedSignal) {
        handlePositionUpdate(obj);
    }
    else if (signalID == parkingStateChangedSignal) {
        handleParkingUpdate(obj);
    }
}

void SimpleApp::handlePositionUpdate(cObject* obj) {
    ChannelMobilityPtrType const mobility = check_and_cast<ChannelMobilityPtrType>(obj);
    curPosition = mobility->getCurrentPosition();
    curSpeed = mobility->getCurrentSpeed();
}

void SimpleApp::handleParkingUpdate(cObject* obj) {
    //this code should only run when used with TraCI
    isParked = mobility->getParkingState();
    if (communicateWhileParked == false) {
        if (isParked == true) {
            (FindModule<BaseConnectionManager*>::findGlobalModule())->unregisterNic(this->getParentModule()->getSubmodule("nic"));
        }
        else {
            Coord pos = mobility->getCurrentPosition();
            (FindModule<BaseConnectionManager*>::findGlobalModule())->registerNic(this->getParentModule()->getSubmodule("nic"), (ChannelAccess*) this->getParentModule()->getSubmodule("nic")->getSubmodule("phy80211p"), &pos);
        }
    }
}

void SimpleApp::handleLowerMsg(cMessage* msg) {

    WaveShortMessage* wsm = dynamic_cast<WaveShortMessage*>(msg);
    ASSERT(wsm);

    if (CAMMessage* cam = dynamic_cast<CAMMessage*>(wsm) ) {
        if(app_debug) std::cerr << myId << ": received CAM: stationId = " << cam->getCam().header.stationID << std::endl;
    } else if (SPATMessage* spat = dynamic_cast<SPATMessage*>(wsm) ) {
        if(app_debug) std::cerr << myId << ": received SPAT: stationId = " << spat->getSpat().header.stationID << std::endl;
    } else {
        error("only CAMMessages and SPATMessages are currently supported");
    }

    delete(msg);
}

void SimpleApp::handleSelfMsg(cMessage* msg) {
    switch (msg->getKind()) {

    default: {
        if (msg)
            DBG_APP << "APP: Error: Got Self Message of unknown kind! Name: " << msg->getName() << endl;
        break;
    }
    }
}

void SimpleApp::finish() {

}

SimpleApp::~SimpleApp() {

    findHost()->unsubscribe(mobilityStateChangedSignal, this);
}



void SimpleApp::sendDown(cMessage* msg) {
    checkAndTrackPacket(msg);
    BaseApplLayer::sendDown(msg);
}

void SimpleApp::sendDelayedDown(cMessage* msg, simtime_t delay) {
    checkAndTrackPacket(msg);
    BaseApplLayer::sendDelayedDown(msg, delay);
}

void SimpleApp::checkAndTrackPacket(cMessage* msg) {
    if (isParked && !communicateWhileParked) error("Attempted to transmit a message while parked, but this is forbidden by current configuration");

    if (dynamic_cast<BasicSafetyMessage*>(msg)) {
        DBG_APP << "sending down a BSM" << std::endl;
        //generatedBSMs++;
    }
    else if (dynamic_cast<WaveServiceAdvertisment*>(msg)) {
        DBG_APP << "sending down a WSA" << std::endl;
        //generatedWSAs++;
    }
    else if (dynamic_cast<WaveShortMessage*>(msg)) {
        DBG_APP << "sending down a wsm" << std::endl;
        //generatedWSMs++;
    }
}

bool SimpleApp::get_is_enabled() {
    return sendBeacons;
}

void SimpleApp::sendBeacon() {
    sendCAM();
    //sendSPAT();
}

void SimpleApp::sendCAM() {
    Enter_Method_Silent();

    CAM_t* cam = generate_CAM();

    CAMMessage* msg = new CAMMessage("CAM");

    msg->setWsmVersion(1);
    msg->setTimestamp(simTime());
    msg->setSenderAddress(myId);
    msg->setRecipientAddress(-1);
    msg->setBitLength(headerLength);
    msg->setPsid(-1);
    msg->setChannelNumber(Channels::CCH);
    msg->addBitLength(beaconLengthBits);
    msg->setUserPriority(beaconPriority);
    msg->setSrcmac(myMacAddress.c_str());

    msg->setCam(*cam);
    track_cam(cam); // cam is deleted there if not needed anymore

    if(first_CAM_generated) {
        if(app_debug) std::cerr << myId << ": sending CAM" << std::endl;
        sendDown(msg);
    } else {
        first_CAM_generated = true;
    }
}

void SimpleApp::sendSPAT() {
    Enter_Method_Silent();

    SPAT_PDU_t* spat = generate_SPAT();

    SPATMessage* msg = new SPATMessage("SPAT");

    msg->setWsmVersion(1);
    msg->setTimestamp(simTime());
    msg->setSenderAddress(myId);
    msg->setRecipientAddress(-1);
    msg->setBitLength(headerLength);
    msg->setPsid(-1);
    msg->setChannelNumber(Channels::CCH);
    msg->addBitLength(beaconLengthBits);
    msg->setUserPriority(beaconPriority);
    msg->setSrcmac(myMacAddress.c_str());

    msg->setSpat(*spat);
    track_spat(spat); // cam is deleted there if not needed anymore

    if(first_SPAT_generated) {
        if(app_debug) std::cerr << myId << ": sending SPAT" << std::endl;
        sendDown(msg);
    } else {
        first_SPAT_generated = true;
    }
}

void SimpleApp::track_cam(CAM_t* cam) {

    if(sent_cams.size() < sent_cams_max_size) {
        sent_cams.push_back(cam);
    } else {
        delete sent_cams.front();
        sent_cams.pop_front();
    }
}

void SimpleApp::track_spat(SPAT_PDU_t* spat) {

    if(sent_spats.size() < sent_spats_max_size) {
        sent_spats.push_back(spat);
    } else {
        delete sent_spats.front();
        sent_spats.pop_front();
    }
}

CAM_t* SimpleApp::generate_CAM() {

    /*
     * Generate a CAM
     */
    CAM_t* cam = static_cast<CAM_t*>(calloc(1, sizeof(CAM_t)));

    // Fill Header
    cam->header.messageID = messageID_cam;
    cam->header.protocolVersion = 1; // was protocolVersion_currentVersion;
    cam->header.stationID = myId; // FIXME: use fitting stationID here

    // Fill CAM
    int64_t currTime = std::chrono::high_resolution_clock::now().time_since_epoch() / std::chrono::nanoseconds(1);
    if (mLastSentCamInfo.timestamp) {
        cam->cam.generationDeltaTime = (currTime - mLastSentCamInfo.timestamp) / (1000000);
    } else {
        cam->cam.generationDeltaTime = 0;
    }
    mLastSentCamInfo.timestamp = currTime;

    // Fill basic Container
    cam->cam.camParameters.basicContainer.stationType = StationType_passengerCar; // FIXME: use fitting stationType here

    if (true /* we always have valid GPS information */) {
        double lat = traci->getLonLat(mobility->getCurrentPosition()).second;
        double lon = traci->getLonLat(mobility->getCurrentPosition()).first;
        double alt = 0; // FIXME we do not have altitude infofmation
        cam->cam.camParameters.basicContainer.referencePosition.latitude = lat * 10000000; // in one-tenth of microdegrees
        cam->cam.camParameters.basicContainer.referencePosition.longitude = lon * 10000000; // in one-tenth of microdegrees
        cam->cam.camParameters.basicContainer.referencePosition.altitude.altitudeValue = alt;
        double currentHeading = getHeading(mLastSentCamInfo.latitude, mLastSentCamInfo.longitude, lat, lon, 0.30); // FIXME use fitting values for mThresholdRadiusForHeading (0.30)

        mLastSentCamInfo.lastHeading = currentHeading;
    } else {
        cam->cam.camParameters.basicContainer.referencePosition.latitude = Latitude_unavailable;
        cam->cam.camParameters.basicContainer.referencePosition.longitude = Longitude_unavailable;
        cam->cam.camParameters.basicContainer.referencePosition.altitude.altitudeValue = AltitudeValue_unavailable;
    }
    cam->cam.camParameters.basicContainer.referencePosition.altitude.altitudeConfidence = AltitudeConfidence_unavailable;

    cam->cam.camParameters.basicContainer.referencePosition.positionConfidenceEllipse.semiMajorConfidence = 0;
    cam->cam.camParameters.basicContainer.referencePosition.positionConfidenceEllipse.semiMajorOrientation = 0;
    cam->cam.camParameters.basicContainer.referencePosition.positionConfidenceEllipse.semiMinorConfidence = 0;

    // Fill High frequency container
    // Could be basic vehicle or RSU and have corresponding details
    if(false /* FIXME we are never an RSU */) {
        cam->cam.camParameters.highFrequencyContainer.present = HighFrequencyContainer_PR_rsuContainerHighFrequency;
        // Optional fields in CAM from RSU
        // cam->cam.camParameters.highFrequencyContainer.choice.rsuContainerHighFrequency.protectedCommunicationZonesRSU
    } else {
        cam->cam.camParameters.highFrequencyContainer.present = HighFrequencyContainer_PR_basicVehicleContainerHighFrequency;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.curvature.curvatureValue = CurvatureValue_unavailable;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.curvature.curvatureConfidence = CurvatureConfidence_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.curvatureCalculationMode = CurvatureCalculationMode_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.driveDirection = DriveDirection_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.heading.headingValue = HeadingValue_unavailable;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.heading.headingConfidence = HeadingConfidence_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.longitudinalAcceleration.longitudinalAccelerationValue = LongitudinalAccelerationValue_unavailable;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.longitudinalAcceleration.longitudinalAccelerationConfidence = AccelerationConfidence_unavailable;

        if (true /* we always have valid OBD2 information */) {
            cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.speed.speedValue = mobility->getCurrentSpeed().length(); // FIXME use correct speed here
        } else {
            cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.speed.speedValue = SpeedValue_unavailable;
        }
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.speed.speedConfidence = SpeedConfidence_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.vehicleLength.vehicleLengthValue = VehicleLengthValue_unavailable;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.vehicleLength.vehicleLengthConfidenceIndication = VehicleLengthConfidenceIndication_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.vehicleWidth = VehicleWidth_unavailable;

        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.yawRate.yawRateValue = YawRateValue_unavailable;
        cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.yawRate.yawRateConfidence = YawRateConfidence_unavailable;

    }


    // Optional
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.accelerationControl->
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.cenDsrcTollingZone->
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.verticalAcceleration
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.steeringWheelAngle
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.performanceClass
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.lanePosition->
    //  cam->cam.camParameters.highFrequencyContainer.choice.basicVehicleContainerHighFrequency.lateralAcceleration

    // Optional part
    //  cam->cam.camParameters.lowFrequencyContainer->present =
    //  cam->cam.camParameters.specialVehicleContainer->present =

    // Printing the cam structure
    //  asn_fprint(stdout, &asn_DEF_CAM, cam);


    //TODO: Free the allocated structure for cam. Is this enough?
    //asn_DEF_CAM.free_struct(&asn_DEF_CAM, cam, 0);
    return cam;
}

double SimpleApp::getDistance(double lat1, double lon1, double lat2, double lon2) {
    double R = 6371; // km
    double dLat = (lat2-lat1) * M_PI/180.0;     //convert to rad
    double dLon = (lon2-lon1) * M_PI/180.0;
    lat1 = lat1 * M_PI/180.0;
    lat2 = lat2 * M_PI/180.0;

    double a = sin(dLat/2) * sin(dLat/2) + sin(dLon/2) * sin(dLon/2) * cos(lat1) * cos(lat2);
    double c = 2 * atan2(sqrt(a), sqrt(1-a));
    return R * c * 1000;                        //convert to m
}

double SimpleApp::getHeading(double lat1, double lon1, double lat2, double lon2, double mThresholdRadiusForHeading) {
    if (getDistance(lat1, lon1, lat2, lon2) < mThresholdRadiusForHeading) {
        //mLogger->logDebug("Ignore heading: not moved more than " +to_string(mThresholdRadiusForHeading) + " meters.");
        return -1;
    }

    double dLat = (lat2-lat1) * M_PI/180.0;     //convert to rad
    double dLon = (lon2-lon1) * M_PI/180.0;
    lat1 = lat1 * M_PI/180.0;

    double phi = atan2(sin(lat1)*dLon, dLat);   //between -pi and +pi
    phi = phi * 180.0/M_PI;                     //convert to deg (-180, +180)
    if(phi < 0) {
        phi += 360;                             //between 0 and 360 deg
    }
    return phi;
}

SPAT_PDU_t* SimpleApp::generate_SPAT() {
    //mLogger->logDebug("Generating CAM as per UPER");
    SPAT_PDU_t* spat = static_cast<SPAT_PDU_t*>(calloc(1, sizeof(SPAT_PDU_t)));
    //if (!cam) {
    //    throw runtime_error("could not allocate CAM_t");
    //}
    // ITS pdu header

    spat->header.stationID = myId;// mIdCounter; //
    spat->header.messageID = messageID_spatem;
    spat->header.protocolVersion = 1; // wasItsPduHeader__protocolVersion_currentVersion;

    //spat->spat.timeStamp = 0;

    //RegionalExtension re;
    //ASN_SEQUENCE_ADD(&spat->spat.regional->list, re);
    //spat->spat.

    IntersectionState* intstate = static_cast<IntersectionState*>(calloc(1, sizeof(IntersectionState)));
    if(intstate==NULL) {
        perror("error!");
    }
    intstate->id.id = myId;
    intstate->revision = 3;

    // IntersectionStatusObject_t
    intstate->status.buf = static_cast<uint8_t*>(calloc(2, sizeof(uint8_t)));
    if(intstate->status.buf==NULL) {
        perror("error!");
    }
    intstate->status.size = 2;
    intstate->status.buf[0] = 1;
    intstate->status.buf[1] = 2;
    intstate->status.bits_unused = 0;
    // IntersectionStatusObject_t

    // MovementState
    MovementState* mov = static_cast<MovementState*>(calloc(1, sizeof(MovementState)));
    mov->signalGroup = myId;//27;

    MovementEvent* evt = static_cast<MovementEvent*>(calloc(1, sizeof(MovementEvent)));

    ASN_SEQUENCE_ADD(&mov->state_time_speed.list, evt);

    ASN_SEQUENCE_ADD(&intstate->states.list, mov);

    ASN_SEQUENCE_ADD(&spat->spatData.intersections.list, intstate);

    return spat;
}
