// This file is part of Lanradio.
//
// Lanradio is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Lanradio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Lanradio.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
// Florian Klingler <klingler@ccs-labs.org>

#ifndef SIMPLEAPP_H_
#define SIMPLEAPP_H_

#include "veins/modules/lanradio/asn1/build/CAM.h"
#include "veins/modules/lanradio/asn1/build/SPAT-PDU.h"
#include "veins/modules/lanradio/LanradioCommunication.h"
#include <map>
#include "veins/base/modules/BaseApplLayer.h"
#include "veins/modules/utility/Consts80211p.h"
#include "veins/modules/messages/WaveShortMessage_m.h"
#include "veins/modules/messages/WaveServiceAdvertisement_m.h"
#include "veins/modules/messages/BasicSafetyMessage_m.h"
#include "veins/base/connectionManager/ChannelAccess.h"
#include "veins/modules/mac/ieee80211p/WaveAppToMac1609_4Interface.h"
#include "veins/modules/mobility/traci/TraCIMobility.h"
#include "veins/modules/mobility/traci/TraCICommandInterface.h"
#include "veins/modules/messages/lanradio/CAMMessage_m.h"
#include "veins/modules/application/lanradio/AppInterface.h"
#include "veins/modules/messages/lanradio/SPATMessage_m.h"

using Veins::TraCIMobility;
using Veins::TraCICommandInterface;
using Veins::AnnotationManager;
using Veins::TraCIMobilityAccess;
using Veins::AnnotationManagerAccess;
using Veins::LanradioCommunication;
using Veins::LanradioCommunicationAccess;

//#define DBG_APP std::cerr << "[" << simTime().raw() << "] " << getParentModule()->getFullPath() << " "

#ifndef DBG_APP
#define DBG_APP EV
#endif


class SimpleApp : public BaseApplLayer, public AppInterface {

    public:
        ~SimpleApp();
        virtual void initialize(int stage);
        virtual void finish();

        virtual void receiveSignal(cComponent* source, simsignal_t signalID, cObject* obj, cObject* details);

        bool get_is_enabled();
        void sendBeacon();
        void sendCAM();
        void sendSPAT();

    protected:

        static const simsignalwrap_t mobilityStateChangedSignal;
        static const simsignalwrap_t parkingStateChangedSignal;

        /** @brief handle messages from below and calls the onWSM, onBSM, and onWSA functions accordingly */
        virtual void handleLowerMsg(cMessage* msg);

        /** @brief handle self messages */
        virtual void handleSelfMsg(cMessage* msg);

        /** @brief sets all the necessary fields in the WSM, BSM, or WSA. */
        virtual void populateWSM(WaveShortMessage*  wsm, int rcvId=0, int serial=0);

        /** @brief this function is called every time the vehicle receives a position update signal */
        virtual void handlePositionUpdate(cObject* obj);

        /** @brief this function is called every time the vehicle parks or starts moving again */
        virtual void handleParkingUpdate(cObject* obj);

        /** @brief compute a point in time that is guaranteed to be in the correct channel interval plus a random offset
         *
         * @param interval the interval length of the periodic message
         * @param chantype the type of channel, either type_CCH or type_SCH
         */
        //virtual simtime_t computeAsynchronousSendingTime(simtime_t interval, t_channel chantype);

        /**
         * @brief overloaded for error handling and stats recording purposes
         *
         * @param msg the message to be sent. Must be a WSM/BSM/WSA
         */
        virtual void sendDown(cMessage* msg);

        /**
         * @brief overloaded for error handling and stats recording purposes
         *
         * @param msg the message to be sent. Must be a WSM/BSM/WSA
         * @param delay the delay for the message
         */
        virtual void sendDelayedDown(cMessage* msg, simtime_t delay);

        /**
         * @brief helper function for error handling and stats recording purposes
         *
         * @param msg the message to be checked and tracked
         */
        virtual void checkAndTrackPacket(cMessage* msg);

        virtual CAM_t* generate_CAM();
        virtual SPAT_PDU_t* generate_SPAT();

    protected:

        /* pointers ill be set when used with TraCIMobility */
        TraCIMobility* mobility;
        TraCICommandInterface* traci;
        TraCICommandInterface::Vehicle* traciVehicle;

        AnnotationManager* annotations;
        WaveAppToMac1609_4Interface* mac;

        /* support for parking currently only works with TraCI */
        bool isParked;
        bool communicateWhileParked;

        /* BSM (beacon) settings */
        uint32_t beaconLengthBits;
        uint32_t  beaconPriority;

        /* state of the vehicle */
        Coord curPosition;
        Coord curSpeed;
        int myId;
        int mySCH;

        std::deque<CAM_t*> sent_cams;
        std::deque<SPAT_PDU_t*> sent_spats;

        void track_cam(CAM_t* cam);
        long sent_cams_max_size;

        void track_spat(SPAT_PDU_t* spat);
        long sent_spats_max_size;

        struct LastSentCamInfo {
            //bool hasGPS = false;
            //gpsPackage::GPS lastGps;
            //bool hasOBD2 = false;
            //obd2Package::OBD2 lastObd2;
            double lastHeading;
            int64_t timestamp = 0;
            double latitude;
            double longitude;
            double altitude;
        };
        LastSentCamInfo mLastSentCamInfo;
        double getDistance(double lat1, double lon1, double lat2, double lon2);
        double getHeading(double lat1, double lon1, double lat2, double lon2, double mThresholdRadiusForHeading);

        bool first_CAM_generated;
        bool first_SPAT_generated;

        std::string myMacAddress;

        bool app_debug;
        bool sendBeacons;
};

#endif /* SIMPLEAPP_H_ */
